<?php 

interface RedBean_ObjectDatabase {
	
	public function load( $type, $id );
	
	public function store( RedBean_OODBBean $bean );
	
	public function trash( RedBean_OODBBean $bean );
	
	public function batch( $type, $ids );
	
	public function dispense( $type );
}

interface RedBean_Plugin {
} 

interface RedBean_Driver {

	public static function getInstance( $host, $user, $pass, $dbname );
	
	public function GetAll( $sql, $aValues=array() );
	
	public function GetCol( $sql, $aValues=array() );
	
	public function GetCell( $sql, $aValues=array() );
	
	public function GetRow( $sql, $aValues=array() );
	
	public function ErrorNo();
	
	public function Errormsg();
	
	public function Execute( $sql, $aValues=array() );
	
	public function Escape( $str );
	
	public function GetInsertID();
	
	public function Affected_Rows();
	
	public function setDebugMode( $tf );
	
	public function GetRaw();
	
	public function CommitTrans();
	
	public function StartTrans();
	
	public function FailTrans();
}
 

class RedBean_Driver_PDO implements RedBean_Driver {
	private $dsn;
	
	private static $instance;
    
	
    private $debug = false;
    
    
    private $pdo;
    
    
    private $affected_rows;
    
    
    private $rs;
    
    
    private $exc =0;
    
    private $connectInfo = array();
    private $isConnected = false;
    
    public static function getInstance($dsn, $user, $pass, $dbname)
    {
        if(is_null(self::$instance))
        {
            self::$instance = new RedBean_Driver_PDO($dsn, $user, $pass);
            
        }
        return self::$instance;
    }
    
    
    public function __construct($dsn, $user, $pass)
    {
	$this->dsn = $dsn;
        $this->connectInfo = array( "pass"=>$pass, "user"=>$user );
    	
    }
    public function connect() {
        if ($this->isConnected) return;
        $user = $this->connectInfo["user"];
        $pass = $this->connectInfo["pass"];
        
        $this->pdo = new PDO(
                $this->dsn,
                $user,
                $pass,
                array(1002 => 'SET NAMES utf8',
                      PDO::ATTR_ERRMODE => PDO::ERRMODE_EXCEPTION,
                      PDO::ATTR_DEFAULT_FETCH_MODE => PDO::FETCH_ASSOC,
					  )
            );
        $this->isConnected = true;
    }
    
    public function GetAll( $sql, $aValues=array() )
    {
        $this->connect();
		
	    	$this->exc = 0;
    	    if ($this->debug)
	        {
	            echo "<HR>" . $sql.print_r($aValues,1);
	        }
		
		try {
			if (strpos("pgsql",$this->dsn)===0){
				$s = $this->pdo->prepare($sql, array(PDO::PGSQL_ATTR_DISABLE_NATIVE_PREPARED_STATEMENT => true));
			}
			else {
				$s = $this->pdo->prepare($sql);
			}
			$s->execute($aValues);
			$this->rs = $s->fetchAll();
			$rows = $this->rs;
		}catch(PDOException $e) {
			
			
			if (version_compare(PHP_VERSION, '5.3.0', '<')) {
				$x = new RedBean_Exception_SQL( $e->getMessage(), 0);
			}
			else {
				$x = new RedBean_Exception_SQL( $e->getMessage(), 0, $e );
			}
			$x->setSQLState( $e->getCode() );
			throw $x;
		}
		if(!$rows) {
			$rows = array();
		}
		if ($this->debug) {
			if (count($rows) > 0) {
				echo "<br><b style='color:green'>resultset: " . count($rows) . " rows</b>";
			}
		}
        return $rows;
    }
    
    
    public function GetCol($sql, $aValues=array())
    {$this->connect();
    	$this->exc = 0;
		$rows = $this->GetAll($sql,$aValues);
		$cols = array();
		if ($rows && is_array($rows) && count($rows)>0){
			foreach ($rows as $row)
			{
				$cols[] = array_shift($row);
			}
		}
        return $cols;
    }
 
    
    public function GetCell($sql, $aValues=array())
    {$this->connect();
    	$this->exc = 0;
        $arr = $this->GetAll($sql,$aValues);
	    $row1 = array_shift($arr);
	    $col1 = array_shift($row1);
        return $col1;
    }
    
    
    public function GetRow($sql, $aValues=array())
    {$this->connect();
    	$this->exc = 0;
       	$arr = $this->GetAll($sql, $aValues);
        return array_shift($arr);
    }
    
    
    public function ErrorNo()
    {
    	if (!$this->exc) return 0;
    	$infos = $this->pdo->errorInfo();
        return $infos[1];
    }
    
    
    public function Errormsg()
    {
    	if (!$this->exc) return "";
        $infos = $this->pdo->errorInfo();
        return $infos[2];
    }
    
    
    public function Execute( $sql, $aValues=array() )
    {$this->connect();
    	$this->exc = 0;
    	    if ($this->debug)
	        {
	            echo "<HR>" . $sql.print_r($aValues,1);
	        }
			try {
			if (strpos("pgsql",$this->dsn)===0){
				$s = $this->pdo->prepare($sql, array(PDO::PGSQL_ATTR_DISABLE_NATIVE_PREPARED_STATEMENT => true));
			}
			else {
				$s = $this->pdo->prepare($sql);
			}
			$s->execute($aValues);
			$this->affected_rows=$s->rowCount();
			return $this->affected_rows;
			}
			catch(PDOException $e) {
				
				
				if (version_compare(PHP_VERSION, '5.3.0', '<')) {
					$x = new RedBean_Exception_SQL( $e->getMessage(), 0);
				}
				else {
					$x = new RedBean_Exception_SQL( $e->getMessage(), 0, $e );
				}
				
				$x->setSQLState( $e->getCode() );
				throw $x;
				
			}
		
    }
    
    
    public function Escape( $str )
    {
        return substr(substr($this->pdo->quote($str), 1), 0, -1);
    }
    
    
    public function GetInsertID()
    {
        return (int) $this->pdo->lastInsertId();
    }
    
    
    public function Affected_Rows()
    {
        return (int) $this->affected_rows;
    }
    
    
    public function setDebugMode( $tf )
    {
        $this->debug = (bool)$tf;
    }
    
    
    public function GetRaw()
    {
        return $this->rs;
    }
	
	public function StartTrans() {$this->connect();
		$this->pdo->beginTransaction();
	}
	
	public function CommitTrans() {$this->connect();
		$this->pdo->commit();
	}
	
	public function FailTrans() {$this->connect();
		$this->pdo->rollback();
	}
	
	public function getDatabaseType() {$this->connect();
		return $this->pdo->getAttribute(PDO::ATTR_DRIVER_NAME);
	}
	
	public function getDatabaseVersion() {$this->connect();
		return $this->pdo->getAttribute(PDO::ATTR_CLIENT_VERSION);
	}
}
 

class RedBean_OODBBean {
	
	private $__info = NULL;
	
	public function import( $arr, $selection=false ) {
		if (is_string($selection)) $selection = explode(",",$selection);
		foreach($arr as $k=>$v) {
			if ($k != "__info") {
                            if (!$selection || ($selection && in_array($k,$selection))) {
                                $this->$k = $v;
                            }
			}
		}
		return $this;
	}
	
	public function export($meta = false) {
		$arr = array();
		foreach($this as $p=>$v) {
			if ($p != "__info" || $meta) {
				$arr[ $p ] = $v;
			}
		}
		return $arr;
	}
	
	public function __get( $property ) {
		if (!isset($this->$property) || $this->$property=="") {
			return NULL;
		}
		return $this->$property;
	}
	
	
	public function __set( $property, $value ) {
		$this->setMeta("tainted",true);
		$this->$property = $value;
	}
	
	public function getMeta( $path, $default = NULL) {
		$ref = $this->__info;
		$parts = explode(".", $path);
		foreach($parts as $part) {
			if (isset($ref[$part])) {
				$ref = $ref[$part];
			}
			else {
				return $default;
			}
		}
		return $ref;
	}
	
	public function setMeta( $path, $value ) {
		$ref = &$this->__info;
		$parts = explode(".", $path);
		$lastpart = array_pop( $parts );
		foreach($parts as $part) {
			if (!isset($ref[$part])) {
				$ref[$part] = array();
			}
			$ref = &$ref[$part];
		}
		$ref[$lastpart] = $value;
	}
	
	public function copyMetaFrom( RedBean_OODBBean $bean ) {
		$this->__info = $bean->__info;
		return $this;
	}
}


abstract class RedBean_Observable {
	
	private $observers = array();
	
	public function addEventListener( $eventname, RedBean_Observer $observer ) {
		if (!isset($this->observers[ $eventname ])) {
			$this->observers[ $eventname ] = array();
		}
		$this->observers[ $eventname ][] = $observer;
	}
	
	public function signal( $eventname, $info ) {
		if (!isset($this->observers[ $eventname ])) {
			$this->observers[ $eventname ] = array();
		}
		foreach($this->observers[$eventname] as $observer) {
			$observer->onEvent( $eventname, $info );
		}
	}
}

interface RedBean_Observer {
	
	
	public function onEvent( $eventname, $bean );
}

interface RedBean_Adapter {
	
	public function getSQL();
	
	public function escape( $sqlvalue );
	
	public function exec( $sql , $aValues=array(), $noevent=false);
	
	public function get( $sql, $aValues = array() );
	
	public function getRow( $sql, $aValues = array() );
	
	public function getCol( $sql, $aValues = array() );
	
	public function getCell( $sql, $aValues = array() );
	
	public function getAssoc( $sql, $values = array() );
	
	public function getInsertID();
	
	public function getAffectedRows();
	
	public function getDatabase();
	
	public function getErrorMsg();
	
	public function startTransaction();
	
	public function commit();
	
	public function rollback();
} 

class RedBean_Adapter_DBAdapter extends RedBean_Observable implements RedBean_Adapter {
	
	private $db = null;
	
	private $sql = "";
	
	public function __construct($database) {
		$this->db = $database;
	}
	
	public function getSQL() {
		return $this->sql;
	}
	
	public function escape( $sqlvalue ) {
		return $this->db->Escape($sqlvalue);
	}
	
	public function exec( $sql , $aValues=array(), $noevent=false) {
		if (!$noevent) {
			$this->sql = $sql;
			$this->signal("sql_exec", $this);
		}
		return $this->db->Execute( $sql, $aValues );
	}
	
	public function get( $sql, $aValues = array() ) {
		$this->sql = $sql;
		$this->signal("sql_exec", $this);
		return $this->db->GetAll( $sql,$aValues );
	}
	
	public function getRow( $sql, $aValues = array() ) {
		$this->sql = $sql;
		$this->signal("sql_exec", $this);
		return $this->db->GetRow( $sql,$aValues );
	}
	
	public function getCol( $sql, $aValues = array() ) {
		$this->sql = $sql;
		$this->signal("sql_exec", $this);
		return $this->db->GetCol( $sql,$aValues );
	}
	
	public function getAssoc( $sql, $aValues = array() ) {
		$this->sql = $sql;
		$this->signal("sql_exec", $this);
		$rows = $this->db->GetAll( $sql, $aValues );
		$assoc = array();
		if ($rows) {
			foreach($rows as $row) {
				if (count($row)>0) {
					$key = array_shift($row);
				}
				if (count($row)>0) {
					$value = array_shift($row);
				}
				else {
					$value = $key;
				}
				$assoc[ $key ] = $value;
			}
		}
		return $assoc;
	}
	
	public function getCell( $sql, $aValues = array() ) {
		$this->sql = $sql;
		$this->signal("sql_exec", $this);
		$arr = $this->db->getCol( $sql, $aValues );
		if ($arr && is_array($arr))	return ($arr[0]); else return false;
	}
	
	public function getInsertID() {
		return $this->db->getInsertID();
	}
	
	public function getAffectedRows() {
		return $this->db->Affected_Rows();
	}
	
	public function getDatabase() {
		return $this->db;
	}
	
	public function getErrorMsg() {
		return $this->db->Errormsg();
	}
	
	public function startTransaction() {
		return $this->db->StartTrans();
	}
	
	public function commit() {
		return $this->db->CommitTrans();
	}
	
	public function rollback() {
		return $this->db->FailTrans();
	}
}
 

interface RedBean_QueryWriter {
	
	const C_SQLSTATE_NO_SUCH_TABLE = 1;
	
	const C_SQLSTATE_NO_SUCH_COLUMN = 2;
	
	
	const C_SQLSTATE_INTEGRITY_CONSTRAINT_VIOLATION = 3;
	
	public function getTables();
	
    public function createTable( $table );
	
    public function getColumns( $table );
	
	public function scanType( $value );
	
    public function addColumn( $table, $column, $type );
	
    public function code( $typedescription );
	
    public function widenColumn( $table, $column, $type );
	
    public function updateRecord( $table, $updatevalues, $id);
	
    public function insertRecord( $table, $insertcolumns, $insertvalues );
	
    public function selectRecord($type, $ids);
	
    public function deleteRecord( $table, $id );
	
    public function addUniqueIndex( $table,$columns );
	
	public function getIDField( $table );
	
	public function selectByCrit( $select, $table, $column, $value, $withUnion=false );
	
	public function deleteByCrit( $table, $crits );
	
	public function noKW($str);
	
	public function sqlStateIn( $state, $list );
}

class RedBean_QueryWriter_MySQL implements RedBean_QueryWriter {
	
	
	const C_DATATYPE_BOOL = 0;
	
	const C_DATATYPE_UINT8 = 1;
	
	const C_DATATYPE_UINT32 = 2;
	
	const C_DATATYPE_DOUBLE = 3;
	
	const C_DATATYPE_TEXT8 = 4;
	
	const C_DATATYPE_TEXT16 = 5;
	
	const C_DATATYPE_TEXT32 = 6;
	
	const C_DATATYPE_SPECIFIED = 99;
	
	public $typeno_sqltype = array(
	RedBean_QueryWriter_MySQL::C_DATATYPE_BOOL=>" SET('1') ",
	RedBean_QueryWriter_MySQL::C_DATATYPE_UINT8=>" TINYINT(3) UNSIGNED ",
	RedBean_QueryWriter_MySQL::C_DATATYPE_UINT32=>" INT(11) UNSIGNED ",
	RedBean_QueryWriter_MySQL::C_DATATYPE_DOUBLE=>" DOUBLE ",
	RedBean_QueryWriter_MySQL::C_DATATYPE_TEXT8=>" VARCHAR(255) ",
	RedBean_QueryWriter_MySQL::C_DATATYPE_TEXT16=>" TEXT ",
	RedBean_QueryWriter_MySQL::C_DATATYPE_TEXT32=>" LONGTEXT "
	);
	
	public $sqltype_typeno = array(
	"set('1')"=>RedBean_QueryWriter_MySQL::C_DATATYPE_BOOL,
	"tinyint(3) unsigned"=>RedBean_QueryWriter_MySQL::C_DATATYPE_UINT8,
	"int(11) unsigned"=>RedBean_QueryWriter_MySQL::C_DATATYPE_UINT32,
	"double" => RedBean_QueryWriter_MySQL::C_DATATYPE_DOUBLE,
	"varchar(255)"=>RedBean_QueryWriter_MySQL::C_DATATYPE_TEXT8,
	"text"=>RedBean_QueryWriter_MySQL::C_DATATYPE_TEXT16,
	"longtext"=>RedBean_QueryWriter_MySQL::C_DATATYPE_TEXT32
	);
	
	public $dtypes = array(
	"booleanset","tinyintus","intus","doubles","varchar255","text","ltext"
	);
	
	protected $adapter;
	
	protected $idfield = "id";
	
	public function getIDField( $type ) {
		return  $this->idfield;
	}
	
	public function check($table) {
		if (strpos($table,"`")!==false) throw new RedBean_Exception_Security("Illegal chars in table name");
		return $this->adapter->escape($table);
	}
	
	public function __construct( RedBean_Adapter $adapter, $frozen = false ) {
		$this->adapter = $adapter;
	}
	
	public function getTables() {
		return $this->adapter->getCol( "show tables" );
	}
	
	public function createTable( $table ) {
		$idfield = $this->getIDfield($table);
		$table = $this->check($table);
		$sql = "
                     CREATE TABLE `$table` (
                    `$idfield` INT( 11 ) UNSIGNED NOT NULL AUTO_INCREMENT ,
                     PRIMARY KEY ( `$idfield` )
                     ) ENGINE = InnoDB DEFAULT CHARSET=utf8 COLLATE=utf8_unicode_ci
			";
		$this->adapter->exec( $sql );
	}
	
	public function getColumns( $table ) {
		$table = $this->check($table);
		$columnsRaw = $this->adapter->get("DESCRIBE `$table`");
		foreach($columnsRaw as $r) {
			$columns[$r["Field"]]=$r["Type"];
		}
		return $columns;
	}
	
	public function scanType( $value ) {
		if (is_null($value)) {
			return RedBean_QueryWriter_MySQL::C_DATATYPE_BOOL;
		}
		$orig = $value;
		$value = strval($value);
		if ($value=="1" || $value=="" || $value=="0") {
			return RedBean_QueryWriter_MySQL::C_DATATYPE_BOOL;
		}
		if (is_numeric($value) && (floor($value)==$value) && $value >= 0 && $value <= 255 ) {
			return RedBean_QueryWriter_MySQL::C_DATATYPE_UINT8;
		}
		if (is_numeric($value) && (floor($value)==$value) && $value >= 0  && $value <= 4294967295 ) {
			return RedBean_QueryWriter_MySQL::C_DATATYPE_UINT32;
		}
		if (is_numeric($value)) {
			return RedBean_QueryWriter_MySQL::C_DATATYPE_DOUBLE;
		}
		if (strlen($value) <= 255) {
			return RedBean_QueryWriter_MySQL::C_DATATYPE_TEXT8;
		}
		return RedBean_QueryWriter_MySQL::C_DATATYPE_TEXT16;
	}
	
	public function addColumn( $table, $column, $type ) {
		$column = $this->check($column);
		$table = $this->check($table);
		$type=$this->typeno_sqltype[$type];
		$sql = "ALTER TABLE `$table` ADD `$column` $type ";
		$this->adapter->exec( $sql );
	}
	
	public function code( $typedescription ) {
		return ((isset($this->sqltype_typeno[$typedescription])) ? $this->sqltype_typeno[$typedescription] : 99);
	}
	
	public function widenColumn( $table, $column, $type ) {
		$column = $this->check($column);
		$table = $this->check($table);
		$newtype = $this->typeno_sqltype[$type];
		$changecolumnSQL = "ALTER TABLE `$table` CHANGE `$column` `$column` $newtype ";
		$this->adapter->exec( $changecolumnSQL );
	}
	
	public function updateRecord( $table, $updatevalues, $id) {
		$idfield = $this->getIDField($table);
		$sql = "UPDATE `".$this->check($table)."` SET ";
		$p = $v = array();
		foreach($updatevalues as $uv) {
			$p[] = " `".$uv["property"]."` = ? ";
			$v[]=( $uv["value"] );
		}
		$sql .= implode(",", $p ) ." WHERE $idfield = ".intval($id);
		$this->adapter->exec( $sql, $v );
	}
	
	public function insertRecord( $table, $insertcolumns, $insertvalues ) {
	
		$idfield = $this->getIDField($table);
		$table = $this->check($table);
		if (count($insertvalues)>0 && is_array($insertvalues[0]) && count($insertvalues[0])>0) {
			foreach($insertcolumns as $k=>$v) {
				$insertcolumns[$k] = "`".$this->check($v)."`";
			}
			$insertSQL = "INSERT INTO `$table` ( $idfield, ".implode(",",$insertcolumns)." ) VALUES ";
			$pat = "( NULL, ". implode(",",array_fill(0,count($insertcolumns)," ? "))." )";
			$insertSQL .= implode(",",array_fill(0,count($insertvalues),$pat));
			foreach($insertvalues as $insertvalue) {
				foreach($insertvalue as $v) {
					$vs[] = ( $v );
				}
			}
			$this->adapter->exec( $insertSQL, $vs );
			return ($this->adapter->getErrorMsg()=="" ?  $this->adapter->getInsertID() : 0);
		}
		else {
			$this->adapter->exec( "INSERT INTO `$table` ($idfield) VALUES(NULL) " );
			return ($this->adapter->getErrorMsg()=="" ?  $this->adapter->getInsertID() : 0);
		}
	}
	
	public function selectRecord($type, $ids) {
		$idfield = $this->getIDField($type);
		$type=$this->check($type);
		$sql = "SELECT * FROM `$type` WHERE $idfield IN ( ".implode(',', array_fill(0, count($ids), " ? "))." )";
		$rows = $this->adapter->get($sql,$ids);
		return ($rows) ? $rows : NULL;
	}
	
	public function deleteRecord( $table, $id) {
		$table = $this->check($table);
		$this->adapter->exec("DELETE FROM `$table` WHERE `".$this->getIDField($table)."` = ? ",array(strval($id)));
	}
	
	public function addUniqueIndex( $table,$columns ) {
		sort($columns); 
		foreach($columns as $k=>$v) {
			$columns[$k]="`".$this->adapter->escape($v)."`";
		}
		$table = $this->check($table);
		$r = $this->adapter->get("SHOW INDEX FROM `$table`");
		$name = "UQ_".sha1(implode(',',$columns));
		if ($r) {
			foreach($r as $i) {
				if ($i["Key_name"]==$name) {
					return;
				}
			}
		}
		$sql = "ALTER IGNORE TABLE `$table`
                ADD UNIQUE INDEX `$name` (".implode(",",$columns).")";
		$this->adapter->exec($sql);
	}
	
	public function selectByCrit( $select, $table, $column, $value, $withUnion=false ) {
		$select = $this->noKW($this->adapter->escape($select));
		$table = $this->noKW($this->adapter->escape($table));
		$column = $this->noKW($this->adapter->escape($column));
		$value = $this->adapter->escape($value);
		$sql = "SELECT $select FROM $table WHERE $column = ? ";
		$values = array($value);
		if ($withUnion) {
			$sql .= " UNION SELECT $column FROM $table WHERE $select = ? ";
			$values[] = $value;
		}
		return $this->adapter->getCol($sql,$values);
	}
	
	public function deleteByCrit( $table, $crits ) {
		$table = $this->noKW($this->adapter->escape($table));
		$values = array();
		foreach($crits as $key=>$val) {
			$key = $this->noKW($this->adapter->escape($key));
			$values[] = $this->adapter->escape($val);
			$conditions[] = $key ."= ? ";
		}
		$sql = "DELETE FROM $table WHERE ".implode(" AND ", $conditions);
		return (int) $this->adapter->exec($sql, $values);
	}
	
	public function noKW($str) {
		return "`".$str."`";
	}
	public function sqlStateIn($state, $list) {
		$sqlState = "0";
		if ($state == "42S02") $sqlState = RedBean_QueryWriter::C_SQLSTATE_NO_SUCH_TABLE;
		if ($state == "42S22") $sqlState = RedBean_QueryWriter::C_SQLSTATE_NO_SUCH_COLUMN;
		if ($state == "23000") $sqlState = RedBean_QueryWriter::C_SQLSTATE_INTEGRITY_CONSTRAINT_VIOLATION;
		return in_array($sqlState, $list);
	}
	
}

class RedBean_QueryWriter_SQLite implements RedBean_QueryWriter {
    
    private $adapter;
    
    private $idfield = "id";
    
    public function getIDField( $type ) {
        return  $this->idfield;
    }
    
    public function check($table) {
        if (strpos($table,"`")!==false) throw new RedBean_Exception_Security("Illegal chars in table name");
        return $this->adapter->escape($table);
    }
    
    public function __construct( RedBean_Adapter $adapter, $frozen = false ) {
        $this->adapter = $adapter;
    }
    
    public function getTables() {
        return $this->adapter->getCol( "SELECT name FROM sqlite_master
			WHERE type='table' AND name!='sqlite_sequence';" );
    }
    
    public function createTable( $table ) {
        $idfield = $this->getIDfield($table);
        $table = $this->check($table);
        $sql = "
                     CREATE TABLE `$table` ( `$idfield` INTEGER PRIMARY KEY AUTOINCREMENT )
                ";
        $this->adapter->exec( $sql );
    }
    
    public function getColumns( $table ) {
        $table = $this->check($table);
        $columnsRaw = $this->adapter->get("PRAGMA table_info('$table')");
        $columns = array();
        foreach($columnsRaw as $r) {
            $columns[$r["name"]]=1;
        }
        return $columns;
    }
    
    public function scanType( $value ) {
        return 1;
    }
    
    public function addColumn( $table, $column, $type ) {
        $column = $this->check($column);
        $table = $this->check($table);
        $sql = "ALTER TABLE `$table` ADD `$column` ";
        $this->adapter->exec( $sql );
    }
    
    public function code( $typedescription ) {
        return 1;
    }
    
    public function widenColumn( $table, $column, $type ) {
        return true;
    }
    
    public function updateRecord( $table, $updatevalues, $id) {
        $idfield = $this->getIDField($table);
        $sql = "UPDATE `".$this->check($table)."` SET ";
        $p = $v = array();
        foreach($updatevalues as $uv) {
            $p[] = " `".$uv["property"]."` = ? ";
            $v[]=strval( $uv["value"] );
        }
        $sql .= implode(",", $p ) ." WHERE $idfield = ".intval($id);
        $this->adapter->exec( $sql, $v );
    }
    
    public function insertRecord( $table, $insertcolumns, $insertvalues ) {
        
        $idfield = $this->getIDField($table);
        $table = $this->check($table);
        if (count($insertvalues)>0 && is_array($insertvalues[0]) && count($insertvalues[0])>0) {
            foreach($insertcolumns as $k=>$v) {
                $insertcolumns[$k] = "`".$this->check($v)."`";
            }
            $insertSQL = "INSERT INTO `$table` ( $idfield, ".implode(",",$insertcolumns)." ) VALUES ";
            $pat = "( NULL, ". implode(",",array_fill(0,count($insertcolumns)," ? "))." )";
            $insertSQL .= implode(",",array_fill(0,count($insertvalues),$pat));
            foreach($insertvalues as $insertvalue) {
                foreach($insertvalue as $v) {
                    $vs[] = strval( $v );
                }
            }
            $this->adapter->exec( $insertSQL, $vs );
            return ($this->adapter->getErrorMsg()=="" ?  $this->adapter->getInsertID() : 0);
        }
        else {
            $this->adapter->exec( "INSERT INTO `$table` ($idfield) VALUES(NULL) " );
            return ($this->adapter->getErrorMsg()=="" ?  $this->adapter->getInsertID() : 0);
        }
    }
    
    public function selectRecord($type, $ids) {
        $idfield = $this->getIDField($type);
        $type=$this->check($type);
        $sql = "SELECT * FROM `$type` WHERE $idfield IN ( ".implode(',', array_fill(0, count($ids), " ? "))." )";
        $rows = $this->adapter->get($sql,$ids);
        return ($rows && is_array($rows) && count($rows)>0) ? $rows : NULL;
    }
    
    public function deleteRecord( $table, $id) {
        $table = $this->check($table);
        $this->adapter->exec("DELETE FROM `$table` WHERE `".$this->getIDField($table)."` = ? ",array(strval($id)));
    }
    
    public function addUniqueIndex( $table,$columns ) {
        $name = "UQ_".sha1(implode(',',$columns));
        $sql = "CREATE UNIQUE INDEX IF NOT EXISTS $name ON $table (".implode(",",$columns).")";
        $this->adapter->exec($sql);
    }
    public function selectByCrit( $select, $table, $column, $value, $withUnion=false ) {
        $select = $this->noKW($this->adapter->escape($select));
        $table = $this->noKW($this->adapter->escape($table));
        $column = $this->noKW($this->adapter->escape($column));
        $value = $this->adapter->escape($value);
        $sql = "SELECT $select FROM $table WHERE $column = ? ";
        $values = array($value);
        if ($withUnion) {
            $sql .= " UNION SELECT $column FROM $table WHERE $select = ? ";
            $values[] = $value;
        }
        return $this->adapter->getCol($sql,$values);
    }
    public function deleteByCrit( $table, $crits ) {
        $table = $this->noKW($this->adapter->escape($table));
        $values = array();
        foreach($crits as $key=>$val) {
            $key = $this->noKW($this->adapter->escape($key));
            $values[] = $val;
            $conditions[] = $key ."= ? ";
        }
        $sql = "DELETE FROM $table WHERE ".implode(" AND ", $conditions);
        return $this->adapter->exec($sql, $values);
    }
    
    public function noKW($str) {
        return "`".$str."`";
    }
    public function sqlStateIn($state, $list) {
        $sqlState = "0";
        if ($state == "HY000") $sqlState = RedBean_QueryWriter::C_SQLSTATE_NO_SUCH_TABLE;
        if ($state == "23000") $sqlState = RedBean_QueryWriter::C_SQLSTATE_INTEGRITY_CONSTRAINT_VIOLATION;
        return in_array($sqlState, $list);
    }
}

class RedBean_QueryWriter_PostgreSQL implements RedBean_QueryWriter {
	
	const C_DATATYPE_INTEGER = 0;
	
	const C_DATATYPE_DOUBLE = 1;
	
	const C_DATATYPE_TEXT = 3;
	
	
    public $typeno_sqltype = array(
        self::C_DATATYPE_INTEGER=>" integer ",
		self::C_DATATYPE_DOUBLE=>" double precision ",
        self::C_DATATYPE_TEXT=>" text "
    );
	
    public $sqltype_typeno = array(
	"integer"=>self::C_DATATYPE_INTEGER,
    "double precision" => self::C_DATATYPE_DOUBLE,
    "text"=>self::C_DATATYPE_TEXT
    );
    
    
    private $adapter;
	
	public function check($table) {
		if (strpos($table,"`")!==false) throw new Redbean_Exception_Security("Illegal chars in table name");
		return $this->adapter->escape($table);
	}
    
    public function __construct( RedBean_Adapter_DBAdapter $adapter ) {
        $this->adapter = $adapter;
    }
    
    public function getTables() {
        return $this->adapter->getCol( "select table_name from information_schema.tables
where table_schema = 'public'" );
    }
	
    public function createTable( $table ) {
		$table = $this->check($table);
		$sql = "
                     CREATE TABLE \"$table\" (
						id SERIAL PRIMARY KEY
                     );
            ";
        $this->adapter->exec( $sql );
    }
	
    public function getColumns( $table ) {
		$table = $this->check($table);
        $columnsRaw = $this->adapter->get("select column_name, data_type from information_schema.columns where table_name='$table'");
        
		foreach($columnsRaw as $r) {
            $columns[$r["column_name"]]=$r["data_type"];
        }
        return $columns;
    }
	
	public function scanType( $value ) { 
		if (is_integer($value) && $value < 2147483648 && $value > -2147483648) {
			return self::C_DATATYPE_INTEGER;
		}
		elseif( is_double($value) ) {
			return self::C_DATATYPE_DOUBLE;
		}
		else {
			return self::C_DATATYPE_TEXT;
		}
    }
	
    public function addColumn( $table, $column, $type ) {
		$column = $this->check($column);
		$table = $this->check($table);
        $type=$this->typeno_sqltype[$type];
        $sql = "ALTER TABLE \"$table\" ADD $column $type ";
        $this->adapter->exec( $sql );
    }
	
    public function code( $typedescription ) {
        return ((isset($this->sqltype_typeno[$typedescription])) ? $this->sqltype_typeno[$typedescription] : 99);
    }
	
    public function widenColumn( $table, $column, $type ) {
        $column = $this->check($column);
		$table = $this->check($table);
		$newtype = $this->typeno_sqltype[$type];
        $changecolumnSQL = "ALTER TABLE \"$table\" \n\t ALTER COLUMN $column TYPE $newtype ";
        try { $this->adapter->exec( $changecolumnSQL ); }catch(Exception $e){ die($e->getMessage()); }
    }
	
    public function updateRecord( $table, $updatevalues, $id) {
		$sql = "UPDATE \"".$this->adapter->escape($this->check($table))."\" SET ";
		$p = $v = array();
		foreach($updatevalues as $uv) {
			$p[] = " \"".$uv["property"]."\" = ? ";
			$v[]=strval( ( $uv["value"] ) );
		}
		$sql .= implode(",", $p ) ." WHERE id = ".intval($id);
		
		$this->adapter->exec( $sql, $v );
    }
	
    public function insertRecord( $table, $insertcolumns, $insertvalues ) {
		$table = $this->check($table);
        if (count($insertvalues)>0 && is_array($insertvalues[0]) && count($insertvalues[0])>0) {
			foreach($insertcolumns as $k=>$v) {
                $insertcolumns[$k] = "".$this->check($v)."";
            }
			$insertSQL = "INSERT INTO \"$table\" ( id, ".implode(",",$insertcolumns)." ) VALUES ";
			$insertSQL .= "( DEFAULT, ". implode(",",array_fill(0,count($insertcolumns)," ? "))." ) RETURNING id";
			
			$ids = array();
			foreach($insertvalues as $insertvalue) {
				$ids[] = $this->adapter->getCell( $insertSQL, $insertvalue );
			}
			if (count($ids)===1) return array_pop($ids); else	return $ids;
			
        }
        else {
			return $this->adapter->getCell( "INSERT INTO \"$table\" (id) VALUES(DEFAULT) RETURNING id " );
        }
    }
	
    public function selectRecord($type, $ids) {
		$type=$this->check($type);
		$sql = "SELECT * FROM $type WHERE id IN ( ".implode(',', array_fill(0, count($ids), " ? "))." )";
		$rows = $this->adapter->get($sql,$ids);
		return ($rows && is_array($rows) && count($rows)>0) ? $rows : NULL;
    }
	
    public function deleteRecord( $table, $value) {
		$table = $this->check($table);
		$column = "id";
	    $this->adapter->exec("DELETE FROM $table WHERE $column = ? ",array(strval($value)));
    }
	
    public function checkChanges($type, $id, $logid) {
		$type = $this->check($type);
		$id = (int) $id;
		$logid = (int) $logid;
		$num = $this->adapter->getCell("
        SELECT count(*) FROM __log WHERE tbl=\"$type\" AND itemid=$id AND action=2 AND id > $logid");
        if ($num) {
			throw new RedBean_Exception_FailedAccessBean("Locked, failed to access (type:$type, id:$id)");
		}
		$newid = $this->insertRecord("__log",array("action","tbl","itemid"),
		   array(array(2,  $type, $id)));
	    if ($this->adapter->getCell("select id from __log where tbl=:tbl AND id < $newid and id > $logid and action=2 and itemid=$id ",
			array(":tbl"=>$type))){
			throw new RedBean_Exception_FailedAccessBean("Locked, failed to access II (type:$type, id:$id)");
		}
		return $newid;
	}
	
    public function addUniqueIndex( $table,$columns ) {
		sort($columns); 
		foreach($columns as $k=>$v){
			$columns[$k]="".$this->adapter->escape($v)."";
		}
		$table = $this->adapter->escape( $this->check($table) );
        $r = $this->adapter->get("SELECT
									i.relname as index_name
								FROM
									pg_class t,
									pg_class i,
									pg_index ix,
									pg_attribute a
								WHERE
									t.oid = ix.indrelid
									AND i.oid = ix.indexrelid
									AND a.attrelid = t.oid
									AND a.attnum = ANY(ix.indkey)
									AND t.relkind = 'r'
									AND t.relname = '$table'
								ORDER BY  t.relname,  i.relname;");
		
        $name = "UQ_".sha1(implode(',',$columns));
        if ($r) {
            foreach($r as $i) {
                if (strtolower( $i["index_name"] )== strtolower( $name )) {
                    return;
                }
            }
        }
		
        $sql = "ALTER TABLE \"$table\"
                ADD CONSTRAINT $name UNIQUE (".implode(",",$columns).")";
		
        $this->adapter->exec($sql);
    }
	
	public function getIDField( $type ) {
		return  "id";
	}
	public function selectByCrit( $select, $table, $column, $value, $withUnion=false ) {
		$select = $this->noKW($this->adapter->escape($select));
		$table = $this->noKW($this->adapter->escape($table));
		$column = $this->noKW($this->adapter->escape($column));
		$value = $this->adapter->escape($value);
		$sql = "SELECT $select FROM $table WHERE $column = ? ";
		$values = array($value);
		if ($withUnion) {
			$sql .= " UNION SELECT $column FROM $table WHERE $select = ? ";
			$values[] = $value;
		}
		return $this->adapter->getCol($sql,$values);
	}
	public function deleteByCrit( $table, $crits ) {
		$table = $this->noKW($this->adapter->escape($table));
		$values = array();
		foreach($crits as $key=>$val) {
			$key = $this->noKW($this->adapter->escape($key));
			$values[] = $val;
			$conditions[] = $key ."= ? ";
		}
		$sql = "DELETE FROM $table WHERE ".implode(" AND ", $conditions);
		return $this->adapter->exec($sql, $values);
	}
	
	public function noKW($str) {
		return "\"".$str."\"";
	}
	
	public function sqlStateIn($state, $list) {
		$sqlState = "0";
		if ($state == "42P01") $sqlState = RedBean_QueryWriter::C_SQLSTATE_NO_SUCH_TABLE;
		if ($state == "42703") $sqlState = RedBean_QueryWriter::C_SQLSTATE_NO_SUCH_COLUMN;
		return in_array($sqlState, $list);
	}
}

class RedBean_Exception extends Exception{}

class RedBean_Exception_SQL extends Exception {
	
	private $sqlState;
	
	public function getSQLState() {
		return $this->sqlState;
	}
	
	public function setSQLState( $sqlState ) {
		$this->sqlState = $sqlState;
	}
	
	public function __toString() {
		return "[".$this->getSQLState()."] - ".$this->getMessage();
	}
} 

class RedBean_Exception_Security extends RedBean_Exception {} 

class RedBean_Exception_FailedAccessBean extends Exception{}

class RedBean_Exception_NotImplemented extends RedBean_Exception {}

class RedBean_Exception_UnsupportedDatabase extends RedBean_Exception {} 

class RedBean_OODB extends RedBean_Observable implements RedBean_ObjectDatabase {
	
	private $stash = NULL;
	
	private $writer;
	
	private $isFrozen = false;
	
	public function __construct( RedBean_QueryWriter $writer ) {
		$this->writer = $writer;
	}
	
	public function freeze( $tf ) {
		$this->isFrozen = (bool) $tf;
	}
	
	public function isFrozen() {
		return (bool) $this->isFrozen;
	}
	
	public function dispense($type ) {
		$bean = new RedBean_OODBBean();
		$bean->setMeta("type", $type );
                $bean->setMeta("sys.oodb", $this);
                $idfield = $this->writer->getIDField($bean->getMeta("type"));
		$bean->$idfield = 0;
		$this->signal( "dispense", $bean );
		$this->check( $bean );
		$bean->setMeta("tainted",false);
		return $bean;
	}
	
	public function check( RedBean_OODBBean $bean ) {
		$idfield = $this->writer->getIDField($bean->getMeta("type"));
		
		if (!isset($bean->$idfield) || !($bean->getMeta("type"))) {
			throw new RedBean_Exception_Security("Bean has incomplete Meta Information");
		}
		
		$pattern = '/[^abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_]/';
		
		if (preg_match($pattern,$bean->getMeta("type"))) {
			throw new RedBean_Exception_Security("Bean Type is invalid");
		}
		
		foreach($bean as $prop=>$value) {
			if (
			is_array($value) ||
				is_object($value) ||
				strlen($prop)<1 ||
				preg_match($pattern,$prop)
			) {
				throw new RedBean_Exception_Security("Invalid Bean: property $prop  ");
			}
		}
	}
	
	public function tableExists($table) {
	
		$tables = $this->writer->getTables();
		return in_array($table, $tables);
	}
	
	public function store( RedBean_OODBBean $bean ) {
		$this->signal( "update", $bean );
		$this->check($bean);
		
		$table = $bean->getMeta("type");
		$idfield = $this->writer->getIDField($table);
		
		if (!$this->isFrozen && !$this->tableExists($table)) {
			$this->writer->createTable( $table );
		}
		$columns = $this->writer->getColumns($table) ;
		
		$insertvalues = array();
		$insertcolumns = array();
		$updatevalues = array();
		foreach( $bean as $p=>$v) {
			if ($p!=$idfield) {
				if (!$this->isFrozen) {
				
					$typeno = $this->writer->scanType($v);
					
					if (isset($columns[$p])) {
					
						$sqlt = $this->writer->code($columns[$p]);
						if ($typeno > $sqlt) {
						
							$this->writer->widenColumn( $table, $p, $typeno );
						}
					}
					else {
					
						$this->writer->addColumn($table, $p, $typeno);
					}
				}
				
				$insertvalues[] = $v;
				$insertcolumns[] = $p;
				$updatevalues[] = array( "property"=>$p, "value"=>$v );
			}
		}
		if (!$this->isFrozen && ($uniques = $bean->getMeta("buildcommand.unique"))) {
			foreach($uniques as $unique) {
				$this->writer->addUniqueIndex( $table, $unique );
			}
		}
		if ($bean->$idfield) {
			if (count($updatevalues)>0) {
				$this->writer->updateRecord( $table, $updatevalues, $bean->$idfield );
			}
			$bean->setMeta("tainted",false);
			return (int) $bean->$idfield;
		}
		else {
			$id = $this->writer->insertRecord( $table, $insertcolumns, array($insertvalues) );
			$bean->$idfield = $id;
			$bean->setMeta("tainted",false);
			return (int) $id;
		}
	}
	
	public function load($type, $id) {
		$id = intval( $id );
		if ($id < 0) throw new RedBean_Exception_Security("Id less than zero not allowed");
		$bean = $this->dispense( $type );
		if ($this->stash && isset($this->stash[$id])) {
			$row = $this->stash[$id];
		}
		else {
			try { $rows = $this->writer->selectRecord($type,array($id));	}catch(RedBean_Exception_SQL $e ){
				if (
					$this->writer->sqlStateIn($e->getSQLState(),
					array(
						RedBean_QueryWriter::C_SQLSTATE_NO_SUCH_COLUMN,
						RedBean_QueryWriter::C_SQLSTATE_NO_SUCH_TABLE)
					)	
				
				)  {
					$rows = 0;
					if ($this->isFrozen) throw $e; 
				}
				else throw $e;
			}
			if (!$rows) return $this->dispense($type);
			$row = array_pop($rows);
		}
		foreach($row as $p=>$v) {
			
			$bean->$p = $v;
		}
		$this->signal( "open", $bean );
		$bean->setMeta("tainted",false);
		return $bean;
	}
	
	public function trash( RedBean_OODBBean $bean ) {
		$idfield = $this->writer->getIDField($bean->getMeta("type"));
		$this->signal( "delete", $bean );
        $this->check( $bean );
		try{
        $this->writer->deleteRecord( $bean->getMeta("type"), $bean->$idfield );
		}catch(RedBean_Exception_SQL $e){
			if (!$this->writer->sqlStateIn($e->getSQLState(),
				array(
					RedBean_QueryWriter::C_SQLSTATE_NO_SUCH_COLUMN,
					RedBean_QueryWriter::C_SQLSTATE_NO_SUCH_TABLE)
			)) throw $e;
		}
	}
	
	public function batch( $type, $ids ) {
		if (!$ids) return array();
		$collection = array();
		try{
		$rows = $this->writer->selectRecord($type,$ids);
		}catch(RedBean_Exception_SQL $e){
			if (!$this->writer->sqlStateIn($e->getSQLState(),
				array(
					RedBean_QueryWriter::C_SQLSTATE_NO_SUCH_COLUMN,
					RedBean_QueryWriter::C_SQLSTATE_NO_SUCH_TABLE)
			)) throw $e;
			$rows = false;
		}
		$this->stash = array();
		if (!$rows) return array();
		foreach($rows as $row) {
			$this->stash[$row[$this->writer->getIDField($type)]] = $row;
		}
		foreach($ids as $id) {
			$collection[ $id ] = $this->load( $type, $id );
		}
		$this->stash = NULL;
		return $collection;
	}
	
	public function convertToBeans($type, $rows) {
		$collection = array();
		$this->stash = array();
		foreach($rows as $row) {
			$id = $row[$this->writer->getIDField($type)];
			$this->stash[$id] = $row;
			$collection[ $id ] = $this->load( $type, $id );
		}
		$this->stash = NULL;
		return $collection;
	}
}


class RedBean_ToolBox {
	
	private $oodb;
	
	private $writer;
	
	private $adapter;
	
	public function __construct( RedBean_OODB $oodb, RedBean_Adapter $adapter, RedBean_QueryWriter $writer ) {
		$this->oodb = $oodb;
		$this->adapter = $adapter;
		$this->writer = $writer;
		return $this;
	}
	
	public function getWriter() {
		return $this->writer;
	}
	
	public function getRedBean() {
		return $this->oodb;
	}
	
	public function getDatabaseAdapter() {
		return $this->adapter;
	}
}

class RedBean_CompatManager {
	
	const C_SYSTEM_MYSQL		= "mysql";
	const C_SYSTEM_SQLITE		= "sqlite";
	const C_SYSTEM_DB2			= "db2";
	const C_SYSTEM_POSTGRESQL	= "pgsql";
	const C_SYSTEM_ORACLE		= "oracle";
	const C_SYSTEM_MSSQL		= "mssql";
	const C_SYSTEM_HYPERTABLE	= "hypertable";
	const C_SYSTEM_INFORMIX		= "informix";
	const C_SYSTEM_SYBASE		= "sybase";
	const C_SYSTEM_FOXPRO		= "foxpro";
	
	private static $ignoreVersion = false;
	
	protected $messageUnsupported = "
Unfortunately ##YOU## is not supported by this module or class.
Supported System(s): ##DBS##.
To suppress this Exception use: RedBean_CompatManager::ignore(TRUE); ";
	
	protected $supportedSystems = array();
	
	public static function ignore( $tf = TRUE ) {
		self::$ignoreVersion = (bool) $tf;
	}
	
	public function scanToolBox( RedBean_ToolBox $toolbox ) {
		
		$brand = strtolower(trim($toolbox->getDatabaseAdapter()->getDatabase()->getDatabaseType()));
		
		
		$version = $toolbox->getDatabaseAdapter()->getDatabase()->getDatabaseVersion();
		if (!is_numeric($version)) {
			$version = 999; 
		}
		
		if (isset($this->supportedSystems[$brand])
			&& ((float)$this->supportedSystems[$brand] <= (float) $version)
		) {
			return true;
		}
		else {
			if (!self::$ignoreVersion) {
				$this->messageUnsupported = str_replace("##YOU##",$brand." v".$version,$this->messageUnsupported);
				$list = array();
				foreach($this->supportedSystems as $supported=>$version) {
					$list[] = " ".$supported . " v".$version."+";
				}
				$this->messageUnsupported = str_replace("##DBS##",implode(",",$list),$this->messageUnsupported);
				throw new RedBean_Exception_UnsupportedDatabase($this->messageUnsupported);
			}
			else {
				return false;
			}
		}
	}
	
	public static function scanDirect( RedBean_ToolBox $toolbox, $list = array() ) {
		$compat = new RedBean_CompatManager();
		$compat->supportedSystems = $list;
		return $compat->scanToolBox($toolbox);
	}
}

class RedBean_AssociationManager extends RedBean_CompatManager {
	
	protected $supportedSystems = array(
		RedBean_CompatManager::C_SYSTEM_MYSQL => "5",
		RedBean_CompatManager::C_SYSTEM_SQLITE=>"3",
		RedBean_CompatManager::C_SYSTEM_POSTGRESQL=>"8"
	);
	
	protected $oodb;
	
	protected $adapter;
	
	protected $writer;
	
	public function __construct( RedBean_ToolBox $tools ) {
		$this->oodb = $tools->getRedBean();
		$this->adapter = $tools->getDatabaseAdapter();
		$this->writer = $tools->getWriter();
	}
	
	public function getTable( $types ) {
		sort($types);
		return implode("_", $types);
	}
	
	public function associate(RedBean_OODBBean $bean1, RedBean_OODBBean $bean2) {
		$table = $this->getTable( array($bean1->getMeta("type") , $bean2->getMeta("type")) );
		$bean = $this->oodb->dispense($table);
		return $this->associateBeans( $bean1, $bean2, $bean );
	}
	
	
	protected function associateBeans(RedBean_OODBBean $bean1, RedBean_OODBBean $bean2, RedBean_OODBBean $bean) {
		
		$idfield1 = $this->writer->getIDField($bean1->getMeta("type"));
		$idfield2 = $this->writer->getIDField($bean2->getMeta("type"));
		
		$property1 = $bean1->getMeta("type") . "_id";
		$property2 = $bean2->getMeta("type") . "_id";
		if ($property1==$property2) $property2 = $bean2->getMeta("type")."2_id";
		$bean->setMeta( "buildcommand.unique" , array( array( $property1, $property2 )));
		$this->oodb->store($bean1);
		$this->oodb->store($bean2);
		$bean->$property1 = $bean1->$idfield1;
		$bean->$property2 = $bean2->$idfield2;
		try {
			return $this->oodb->store( $bean );
		}
		catch(RedBean_Exception_SQL $e)  {
			if (!$this->writer->sqlStateIn($e->getSQLState(),
				array(
					RedBean_QueryWriter::C_SQLSTATE_INTEGRITY_CONSTRAINT_VIOLATION
			))) throw $e;
		}
	}
	
	
	public function related( RedBean_OODBBean $bean, $type, $getLinks=false ) {
		$table = $this->getTable( array($bean->getMeta("type") , $type) );
		$idfield = $this->writer->getIDField($bean->getMeta("type"));
		if ($type==$bean->getMeta("type")) {
			$type .= "2";
			$cross = 1;
		}
		else $cross=0;
		if (!$getLinks) $targetproperty = $type."_id"; else $targetproperty="id";
		$property = $bean->getMeta("type")."_id";
		try {
			if ($cross) {
				$sqlFetchKeys = $this->writer->selectByCrit(
					$targetproperty,
					$table,
					$property,
					$bean->$idfield,
					true
				);
			}
			else {
				$sqlFetchKeys = $this->writer->selectByCrit(
					$targetproperty,
					$table,
					$property,
					$bean->$idfield
				);
			}
			return ( $sqlFetchKeys );
		}catch(RedBean_Exception_SQL $e){
			
			if (!$this->writer->sqlStateIn($e->getSQLState(),
				array(
					RedBean_QueryWriter::C_SQLSTATE_NO_SUCH_COLUMN,
					RedBean_QueryWriter::C_SQLSTATE_NO_SUCH_TABLE)
			)) throw $e;
			return array();
		}
	}
	
	public function unassociate(RedBean_OODBBean $bean1, RedBean_OODBBean $bean2) {
		$this->oodb->store($bean1);
		$this->oodb->store($bean2);
		$table = $this->getTable( array($bean1->getMeta("type") , $bean2->getMeta("type")) );
		$idfield1 = $this->writer->getIDField($bean1->getMeta("type"));
		$idfield2 = $this->writer->getIDField($bean2->getMeta("type"));
		$type = $bean1->getMeta("type");
		if ($type==$bean2->getMeta("type")) { 
			$type .= "2";
			$cross = 1;
		}
		else $cross = 0;
		$property1 = $type."_id";
		$property2 = $bean2->getMeta("type")."_id";
		$value1 = (int) $bean1->$idfield1;
		$value2 = (int) $bean2->$idfield2;
		try {
			$this->writer->deleteByCrit($table,array($property1=>$value1,$property2=>$value2));
			if ($cross) {
				$this->writer->deleteByCrit($table,array($property2=>$value1,$property1=>$value2));
			}
		}catch(RedBean_Exception_SQL $e){
			if (!$this->writer->sqlStateIn($e->getSQLState(),
				array(
					RedBean_QueryWriter::C_SQLSTATE_NO_SUCH_COLUMN,
					RedBean_QueryWriter::C_SQLSTATE_NO_SUCH_TABLE)
			)) throw $e;
		}
	}
	
	public function clearRelations(RedBean_OODBBean $bean, $type) {
		$this->oodb->store($bean);
		$table = $this->getTable( array($bean->getMeta("type") , $type) );
		$idfield = $this->writer->getIDField($bean->getMeta("type"));
		if ($type==$bean->getMeta("type")) {
			$property2 = $type."2_id";
			$cross = 1;
		}
		else $cross = 0;
		$property = $bean->getMeta("type")."_id";
		try {
			$this->writer->deleteByCrit($table,array($property=>$bean->$idfield));
			if ($cross) {
				$this->writer->deleteByCrit($table,array($property2=>$bean->$idfield));
			}
		}catch(RedBean_Exception_SQL $e){
			if (!$this->writer->sqlStateIn($e->getSQLState(),
				array(
					RedBean_QueryWriter::C_SQLSTATE_NO_SUCH_COLUMN,
					RedBean_QueryWriter::C_SQLSTATE_NO_SUCH_TABLE)
			)) throw $e;
		}
	}
	
	public function set1toNAssoc(RedBean_OODBBean $bean1, RedBean_OODBBean $bean2) {
		$type = $bean1->getMeta("type");
		$this->clearRelations($bean2, $type);
		$this->associate($bean1, $bean2);
		if (count( $this->related($bean2, $type) )===1) {
			return $this;
		}
		else {
			throw new RedBean_Exception_SQL("Failed to enforce 1toN Relation for $type ");
		}
	}
}

class RedBean_TreeManager extends RedBean_CompatManager {
	
	protected $supportedSystems = array(
		RedBean_CompatManager::C_SYSTEM_MYSQL => "5",
		RedBean_CompatManager::C_SYSTEM_SQLITE=>"3"
	);
	
	private $property = "parent_id";
	
	private $oodb;
	
	private $adapter;
	
	private $writer;
	
	public function __construct( RedBean_ToolBox $tools ) {
	
		$this->oodb = $tools->getRedBean();
		$this->adapter = $tools->getDatabaseAdapter();
		$this->writer = $tools->getWriter();
	}
    
    private function equalTypes( RedBean_OODBBean $bean1, RedBean_OODBBean $bean2 ) {
        if ($bean1->getMeta("type")!==$bean2->getMeta("type")) {
            throw new RedBean_Exception_Security("Incompatible types, tree can only work with identical types.");
        }
    }
	
	public function attach( RedBean_OODBBean $parent, RedBean_OODBBean $child ) {
                $this->equalTypes( $parent, $child );
		$idfield = $this->writer->getIDField($parent->getMeta("type"));
		if (!intval($parent->$idfield)) $this->oodb->store($parent);
		$child->{$this->property} = $parent->$idfield;
		$this->oodb->store($child);
	}
	
	public function children( RedBean_OODBBean $parent ) {
		$idfield = $this->writer->getIDField($parent->getMeta("type"));
		try {
			$ids = $this->writer->selectByCrit( $idfield,
				$parent->getMeta("type"),
				$this->property,
				intval( $parent->$idfield ) );
		}
		catch(RedBean_Exception_SQL $e) {
			return array();
		}
		return $this->oodb->batch($parent->getMeta("type"),$ids	);
	}
        public function getParent( RedBean_OODBBean $bean ) {
            return $this->oodb->load( $bean->getMeta("type"), (int)$bean->parent_id);
        }
}

class RedBean_LinkManager extends RedBean_CompatManager {
	
	protected $supportedSystems = array(
		RedBean_CompatManager::C_SYSTEM_MYSQL => "5",
		RedBean_CompatManager::C_SYSTEM_SQLITE=>"3",
		RedBean_CompatManager::C_SYSTEM_POSTGRESQL=>"8"
	);
	
	protected $oodb;
	
	protected $adapter;
	
	protected $writer;
	
	public function __construct( RedBean_ToolBox $tools ) {
		$this->oodb = $tools->getRedBean();
		$this->adapter = $tools->getDatabaseAdapter();
		$this->writer = $tools->getWriter();
	}
	
	public function getLinkField( $typeName ) {
		$fieldName = strtolower( $typeName )."_id";
                $fieldName = preg_replace( "/\W/","", $fieldName );
                return $fieldName;
        }
	
	public function link(RedBean_OODBBean $bean1, RedBean_OODBBean $bean2) {
                if (!$bean2->id) {
                    
                    $this->oodb->store( $bean2 );
                }
		$fieldName = $this->getLinkField( $bean2->getMeta("type") );
                $bean1->$fieldName = $bean2->id;
                return $this;
	}
	
        public function getBean( RedBean_OODBBean $bean, $typeName ) {
            $fieldName = $this->getLinkField($typeName);
            $id = (int)$bean->$fieldName;
            if ($id) {
                return $this->oodb->load($typeName, $id);
            }
            else {
                return $this->oodb->dispense($typeName);
            }
        }
        
        public function breakLink( RedBean_OODBBean $bean, $typeName ) {
            $fieldName = $this->getLinkField($typeName);
            $bean->$fieldName = NULL;
        }
        
        public function getKey(RedBean_OODBBean $bean, $typeName) {
            $fieldName = $this->getLinkField($typeName);
            $id = (int)$bean->$fieldName;
            return $id;
        }
}

class RedBean_ExtAssociationManager extends RedBean_AssociationManager {
	
	public function extAssociate(RedBean_OODBBean $bean1, RedBean_OODBBean $bean2, RedBean_OODBBean $baseBean ) {
		$table = $this->getTable( array($bean1->getMeta("type") , $bean2->getMeta("type")) );
		$baseBean->setMeta("type", $table );
		return $this->associateBeans( $bean1, $bean2, $baseBean );
	}
	
	
} 

class RedBean_Setup {
		
		private static $observers = array();
		
		private static $toolbox = NULL;
		
		private static function checkDSN($dsn) {
			$dsn = trim($dsn);
			$dsn = strtolower($dsn);
			if (
				strpos($dsn, "mysql:")!==0
				&& strpos($dsn,"sqlite:")!==0
				&& strpos($dsn,"pgsql:")!==0
			) {
				
				throw new RedBean_Exception_NotImplemented("
					Support for this DSN has not been implemented yet. \n
					Begin your DSN with: 'mysql:' or 'sqlite:'
				");
			}
			else {
				return true;
			}
		}
		
        public static function kickstart( $dsn, $username, $password, $frozen=false ) {
			self::checkDSN($dsn);
            $pdo = new RedBean_Driver_PDO( $dsn,$username,$password );
            $adapter = new RedBean_Adapter_DBAdapter( $pdo );
			
            if (strpos($dsn,"pgsql")===0) {
                $writer = new RedBean_QueryWriter_PostgreSQL( $adapter, $frozen );
            }
			else if (strpos($dsn,"sqlite")===0){
				 $writer = new RedBean_QueryWriter_SQLite( $adapter, $frozen );
			}
			else {
                $writer = new RedBean_QueryWriter_MySQL( $adapter, $frozen );
			}
            $redbean = new RedBean_OODB( $writer );
			$toolbox = new RedBean_ToolBox( $redbean, $adapter, $writer );
            
            
			self::$toolbox = $toolbox;
            return self::$toolbox;
        }
		
		public static function kickstartDev( $dsn, $username="root", $password="" ) {
			$toolbox = self::kickstart($dsn, $username, $password);
			return $toolbox;
		}
		
		public static function kickstartDevL( $dsn ) {
			self::checkDSN($dsn);
            $pdo = new RedBean_Driver_PDO( $dsn ,"","");
            $adapter = new RedBean_Adapter_DBAdapter( $pdo );
            $writer = new RedBean_QueryWriter_SQLite( $adapter, false );
            $redbean = new RedBean_OODB( $writer );
			$toolbox = new RedBean_ToolBox( $redbean, $adapter, $writer );
            
			self::$toolbox = $toolbox;
            return self::$toolbox;
		}
		
		public static function KickStartDevWithJournal($dsn, $username="root", $password="") {
			$toolbox = self::kickstart($dsn, $username, $password);
			$redbean = $toolbox->getRedBean();
			$logger = new RedBean_Plugin_ChangeLogger( $toolbox );
			self::$observers["logger"] = $logger;
			$redbean->addEventListener( "open", $logger );
			$redbean->addEventListener( "update", $logger);
			$redbean->addEventListener( "delete", $logger);
			return $toolbox;
		}
		
		public static function kickstartFrozen( $dsn, $username, $password ) {
			$toolbox = self::kickstart($dsn, $username, $password, true);
			$toolbox->getRedBean()->freeze(true);
			return $toolbox;
		}
		
		public static function kickstartDebug( $dsn, $username="root", $password="" ) {
			$toolbox = self::kickstart($dsn, $username, $password);
			$toolbox->getDatabaseAdapter()->getDatabase()->setDebugMode( true );
			return $toolbox;
		}
		
		public static function getAttachedObservers() {
			return self::$observers;
		}
		
		public static function getToolBox() {
			return self::$toolbox;
		}
}


class RedBean_SimpleStat extends RedBean_CompatManager implements RedBean_Plugin {
	
	private $oodb;
	
	private $adapter;
	
	private $writer;
	
	protected $supportedSystems = array(
		RedBean_CompatManager::C_SYSTEM_MYSQL => "5",
		RedBean_CompatManager::C_SYSTEM_SQLITE => "3"
	);
	
	public function __construct( RedBean_ToolBox $tools ) {
		
		$this->scanToolBox( $tools );
		$this->oodb = $tools->getRedBean();
		$this->adapter = $tools->getDatabaseAdapter();
		$this->writer = $tools->getWriter();
	}
	
	public function numberOf(RedBean_OODBBean $bean) {
		$type = $this->adapter->escape( $bean->getMeta("type") );
		return (int) $this->adapter->getCell("SELECT count(*) FROM `$type`");
	}
	
}

class RedBean_Plugin_ChangeLogger extends RedBean_CompatManager implements RedBean_Plugin,RedBean_Observer {
	
	protected $supportedSystems = array(
		RedBean_CompatManager::C_SYSTEM_MYSQL => "5",
	);
    
    private $writer;
	
	private $adapter;
	
	private $stash = array();
	
	private $redbean;
	
    public function __construct(RedBean_ToolBox $toolbox) {
		
		$this->scanToolBox( $toolbox );
        $this->writer = $toolbox->getWriter();
		$this->adapter = $toolbox->getDatabaseAdapter();
		$this->redbean = $toolbox->getRedBean();
		if (!$this->redbean->isFrozen()) {
			$this->adapter->exec("
						CREATE TABLE IF NOT EXISTS `__log` (
						`id` INT( 11 ) NOT NULL AUTO_INCREMENT PRIMARY KEY ,
						`tbl` VARCHAR( 255 ) NOT NULL ,
						`action` TINYINT( 2 ) NOT NULL ,
						`itemid` INT( 11 ) NOT NULL
						) ENGINE = MYISAM ;
				"); 
		}
		$maxid = $this->adapter->getCell("SELECT MAX(id) FROM __log");
		$this->adapter->exec("DELETE FROM __log WHERE id < $maxid - 200 ");
    }
	
    public function onEvent( $event, $item ) { 
        $id = $item->id;
        if (! ((int) $id)) $event="open";
        $type = $item->getMeta("type");
        if ($event=="open") {
			if (isset($this->stash[$id])) {
				$insertid = $this->stash[$id];
				unset($this->stash[$id]);
				return $insertid;
			}
			$insertid = $this->writer->insertRecord("__log",array("action","tbl","itemid"),
            array(array(1,  $type, $id)));
			$item->setMeta("opened",$insertid);
			
        }
        if ($event=="update" || $event=="delete") {
            if (($item->getMeta("opened"))) $oldid = $item->getMeta("opened"); else $oldid=0;
            $newid = $this->checkChanges($type,$id, $oldid);
	        $item->setMeta("opened",$newid);
        }
    }
	
	public function preLoad( $type, $ids ) {
		$this->adapter->exec("INSERT INTO __log (id,action,tbl,itemid)
		VALUES(NULL, :action,:tbl,:id)",array(":action"=>1,":tbl"=>"__no_type__",":id"=>0)); 
		$insertid = $this->adapter->getInsertID();
		$values = array();
		foreach($ids as $id) { 
			$this->stash[$id]=$insertid; 
			$values[] = array(1, $type, $id); 
		}
		$this->writer->insertRecord("__log",array("action","tbl","itemid"), $values); 
	}
	
	public function testingOnly_getStash() {
		return $this->stash;
	}
	
    public function checkChanges($type, $id, $logid) {
		$type = $this->writer->check($type);
		$id = (int) $id;
		$logid = (int) $logid;
		$num = $this->adapter->getCell("
        SELECT count(*) FROM __log WHERE tbl=\"$type\" AND itemid=$id AND action=2 AND id > $logid");
        if ($num) {
			throw new RedBean_Exception_FailedAccessBean("Locked, failed to access (type:$type, id:$id)");
		}
		$this->adapter->exec("INSERT INTO __log (id,action,tbl,itemid) VALUES(NULL, 2,:tbl,:id)",array(":tbl"=>$type, ":id"=>$id));
		$newid = $this->adapter->getInsertID();
	    if ($this->adapter->getCell("select id from __log where tbl=:tbl AND id < $newid and id > $logid and action=2 and itemid=$id ",
			array(":tbl"=>$type))){
			throw new RedBean_Exception_FailedAccessBean("Locked, failed to access II (type:$type, id:$id)");
		}
		return $newid;
	}
}

class RedBean_Plugin_Cache extends RedBean_Observable implements RedBean_Plugin, RedBean_ObjectDatabase {
	
	private $oodb;
	
	private $writer;
	
	private $cache = array();
	
	private $originals = array();
	
	private $columnCounter = 0;
	
	public function __construct( RedBean_OODB $oodb, RedBean_ToolBox $toolBox ) {
		$this->oodb = $oodb;
		$this->writer = $toolBox->getWriter();
	}
	
	public function addEventListener($event, RedBean_Observer $o) {
		$this->oodb->addEventListener($event, $o);
	}
	
	private function generateKey( RedBean_OODBBean $bean ) {
		$type=$bean->getMeta("type");
		$idfield = $this->writer->getIDField($type);
		$id = $bean->$idfield;
		return sha1($type."-".$id);
	}
	
	private function putInCache( RedBean_OODBBean $bean ) {
		$key = $this->generateKey($bean);
		$this->cache[$key]=$bean;
		$copy = clone $bean;
		$copy->copyMetaFrom($bean);
		$this->originals[ $key ] = $copy;
		return $this;
	}
	
	private function fetchFromCache( RedBean_OODBBean $bean ) {
		$key = $this->generateKey($bean);
		if (isset($this->cache[$key])) {
			return $this->cache[$key];
		}
		else {
			return NULL;
		}
	}
	
	private function fetchFromCacheByTypeID( $type, $id ) {
		$bean = $this->oodb->dispense($type);
		$idfield = $this->writer->getIDField($type);
		$bean->$idfield = $id;
		return $this->fetchFromCache($bean);
	}
	
	private function fetchOriginal(RedBean_OODBBean $bean) {
		$key = $this->generateKey($bean);
		if (isset($this->originals[$key])) {
			return $this->originals[$key];
		}
		else {
			return NULL;
		}
	}
	
	private function removeFromCache( RedBean_OODBBean $bean ) {
		$key = $this->generateKey($bean);
		unset($this->cache[$key]);
		unset($this->originals[$key]);
		return $this;
	}
	
	public function load( $type, $id ) {
		$bean = $this->fetchFromCacheByTypeID($type, $id);
		if ($bean) {
			return $bean;
		}
		else {
			$bean = $this->oodb->load($type, $id);
			$this->putInCache($bean);
			return $bean;
		}
	}
	
	
	public function store( RedBean_OODBBean $bean ) {
		$this->columnCounter = 0;
		$type=$bean->getMeta("type");
		$idfield = $this->writer->getIDField($type);
		$newbean = $this->oodb->dispense($type);
		$newbean->$idfield = $bean->$idfield;
		$oldBean = $this->fetchOriginal($bean);
		
		if ($oldBean) {
			
			$dirty = false;
			
			foreach($oldBean as $p=>$v) {
				if ($v !== $bean->$p && $p!=$idfield) {
					$newbean->$p = $bean->$p;
					
					$this->columnCounter++; 
					
					$dirty=true;
				}
			}
			
			
			foreach($bean as $p=>$v){
				if (!isset($oldBean->$p)) {
					$dirty=true;
					$newbean->$p = $bean->$p;
					$this->columnCounter++; 
				}
			}
			
			
			if ($dirty) {
				$newbean->copyMetaFrom($bean);
				$id = $this->oodb->store($newbean);
				$bean->copyMetaFrom($newbean);
				$this->putInCache($bean);
				return $id;
			}
			else {
				return $bean->$idfield;
			}
		}
		else {
			$id = $this->oodb->store($bean);
			$this->putInCache($bean);
			return $id;
		}
	}
	
	public function trash( RedBean_OODBBean $bean ) {
		$this->removeFromCache($bean);
		return $this->oodb->trash($bean);
	}
	
	public function batch( $type, $ids ) {
		$idfield = $this->writer->getIDField($type);
		$collect = array();
		foreach($ids as $id) {
			$bean = $this->fetchFromCacheByTypeID($type, $id);
			if ($bean) $collect[$id] = $bean;
		}
		if (count($collect) == count($ids)) { 
			return $collect;
		}
		else {
			$beans = $this->oodb->batch($type, $ids);
			foreach($beans as $bean) {
				$this->putInCache( $bean );
			}
			return $beans;
		}
	}
	
	public function dispense( $type ){
		return $this->oodb->dispense($type);
	}
	
	public function test_getColCount() {
		return $this->columnCounter;
	}
}

class RedBean_Plugin_Finder implements RedBean_Plugin {
	
	
	public static function where( $type, $SQL = " 1 ", $values=array(),
			$tools = false, $ignoreGSQLWarn = false ) {
		if ($SQL==="") $SQL = " 1 ";
		
		$type = preg_replace("/\W/","", $type);
		
		if (!$tools) $tools = RedBean_Setup::getToolBox();
		RedBean_CompatManager::scanDirect($tools, array(
				RedBean_CompatManager::C_SYSTEM_MYSQL => "5",
				RedBean_CompatManager::C_SYSTEM_SQLITE => "3",
				RedBean_CompatManager::C_SYSTEM_POSTGRESQL => "7"
			));
		
		
		$redbean = $tools->getRedBean();
		$adapter = $tools->getDatabaseAdapter();
		$writer = $tools->getWriter();
		
		
		if (!$redbean->isFrozen()) { 
			$SQL = self::parseGoldSQL($SQL, $type, $tools);
		}
		else { 
			if (!$ignoreGSQLWarn && strpos($SQL,"@")!==false) {
				throw new RedBean_Exception_SQL("Gold SQL is
					only allowed in FLUID mode,
					to ignore use extra argument TRUE for Finder::Where");
			}
		}
		
		
		try{
			$SQL = "SELECT * FROM $type WHERE ".$SQL;
			
			$rows = $adapter->get($SQL, $values);
		}
		catch(RedBean_Exception_SQL $e) { 
			if ($writer->sqlStateIn($e->getSQLState(),array(
				RedBean_QueryWriter::C_SQLSTATE_NO_SUCH_COLUMN,
				RedBean_QueryWriter::C_SQLSTATE_NO_SUCH_TABLE
			))){
				return array();
			}
			else {
				throw $e;
			}
		}
		
		
		
		return $redbean->convertToBeans($type, $rows);
		
	}
	
	public static function parseGoldSQL( $SQL, $currentTable,  RedBean_ToolBox $toolbox ) {
		
		$matches = array();
		
		$pattern = "/@[\w\.]+/";
		
		if (preg_match_all($pattern, $SQL, $matches)) {
			
			$columns = array_keys( $toolbox->getWriter()->getColumns($currentTable) );
			
			$tables = $toolbox->getWriter()->getTables();
		
			
			$checks = array_shift( $matches );
			
			foreach($checks as $checkItem) {
				
				$itemName = substr($checkItem, 1);
				
				if (strpos($itemName,".")!==false) { 
					list($table, $column) = explode(".", $itemName);
					if (!in_array($table, $tables)) {
						
						$SQL = str_replace("@".$itemName, "NULL", $SQL);
						
					}
					else {
						$tableCols = array_keys( $toolbox->getWriter()->getColumns($table) );
						if (!in_array($column, ($tableCols))) {
							$SQL = str_replace("@".$itemName, "NULL", $SQL);
						}
						else {
							$SQL = str_replace("@".$itemName, $itemName, $SQL);
						}
					}
				}
				else {
				
					if (!in_array($itemName, ($columns))) {
	
						$SQL = str_replace("@".$itemName, "NULL", $SQL);
						
					}
					else {
						$SQL = str_replace("@".$itemName, $itemName, $SQL);
					}
				}
				
			}
			
			
		}
		return $SQL;
		
	}
}

class Finder extends RedBean_Plugin_Finder { }


class RedBean_Plugin_Constraint {
	
	private static $fkcache = array();
	
	public static function addConstraint( RedBean_OODBBean $bean1, RedBean_OODBBean $bean2, $dontCache = false ) {
		
		$toolbox = RedBean_Setup::getToolBox();
		RedBean_CompatManager::scanDirect($toolbox, array(
			RedBean_CompatManager::C_SYSTEM_MYSQL => "5",
			RedBean_CompatManager::C_SYSTEM_SQLITE => "3",
			RedBean_CompatManager::C_SYSTEM_POSTGRESQL => "7",));
		
		$association = new RedBean_AssociationManager( $toolbox );
		$writer = $toolbox->getWriter();
		$oodb = $toolbox->getRedBean();
		$adapter = $toolbox->getDatabaseAdapter();
		
		if ($oodb->isFrozen()) return false;
		$table1 = $bean1->getMeta("type");
		$table2 = $bean2->getMeta("type");
		$table = $association->getTable( array( $table1,$table2) );
		$idfield1 = $writer->getIDField($bean1->getMeta("type"));
		$idfield2 = $writer->getIDField($bean2->getMeta("type"));
		$bean = $oodb->dispense($table);
		$property1 = $bean1->getMeta("type") . "_id";
		$property2 = $bean2->getMeta("type") . "_id";
		if ($property1==$property2) $property2 = $bean2->getMeta("type")."2_id";
		$table = $adapter->escape($table);
		$table1 = $adapter->escape($table1);
		$table2 = $adapter->escape($table2);
		$property1 = $adapter->escape($property1);
		$property2 = $adapter->escape($property2);
		
		
		$fkCode = "fk".md5($table.$property1.$property2);
		if (isset(self::$fkcache[$fkCode])) return false;
		
		
		try{
			if ($writer instanceof RedBean_QueryWriter_PostgreSQL) {
				return self::constraintPostgreSQL($toolbox, $table, $table1, $table2, $property1, $property2, $dontCache);
			}
			if ($writer instanceof RedBean_QueryWriter_SQLite) {
				return self::constraintSQLite($toolbox, $table, $table1, $table2, $property1, $property2, $dontCache);
			}
			if ($writer instanceof RedBean_QueryWriter_MySQL) {
				return self::constraintMySQL($toolbox, $table, $table1, $table2, $property1, $property2, $dontCache);
			}
		}
		catch(RedBean_Exception_SQL $e){
			if (!$writer->sqlStateIn($e->getSQLState(),
				array(
					RedBean_QueryWriter::C_SQLSTATE_NO_SUCH_COLUMN,
					RedBean_QueryWriter::C_SQLSTATE_NO_SUCH_TABLE)
			)) throw $e;
		}
		
		return false;
	
	}
	
	private static function constraintPostgreSQL($toolbox, $table, $table1, $table2, $property1, $property2, $dontCache) {
			$writer = $toolbox->getWriter();
			$oodb = $toolbox->getRedBean();
			$adapter = $toolbox->getDatabaseAdapter();
			$fkCode = "fk".md5($table.$property1.$property2);
			$sql = "
					SELECT
							c.oid,
							n.nspname,
							c.relname,
							n2.nspname,
							c2.relname,
							cons.conname
					FROM pg_class c
					JOIN pg_namespace n ON n.oid = c.relnamespace
					LEFT OUTER JOIN pg_constraint cons ON cons.conrelid = c.oid
					LEFT OUTER JOIN pg_class c2 ON cons.confrelid = c2.oid
					LEFT OUTER JOIN pg_namespace n2 ON n2.oid = c2.relnamespace
					WHERE c.relkind = 'r'
					AND n.nspname IN ('public')
					AND (cons.contype = 'f' OR cons.contype IS NULL)
					AND
					(  cons.conname = '{$fkCode}a'	OR  cons.conname = '{$fkCode}b' )
			";
			$rows = $adapter->get( $sql );
			if (!count($rows)) {
				if (!$dontCache) self::$fkcache[ $fkCode ] = true;
				$sql1 = "ALTER TABLE $table ADD CONSTRAINT
						{$fkCode}a FOREIGN KEY ($property1)
						REFERENCES $table1 (id) ON DELETE CASCADE ";
				$sql2 = "ALTER TABLE $table ADD CONSTRAINT
						{$fkCode}b FOREIGN KEY ($property2)
						REFERENCES $table2 (id) ON DELETE CASCADE ";
				$adapter->exec($sql1);
				$adapter->exec($sql2);
			}
			return true;
	}
	
	private static function constraintMySQL($toolbox, $table, $table1, $table2, $property1, $property2, $dontCache) {
		$writer = $toolbox->getWriter();
		$oodb = $toolbox->getRedBean();
		$adapter = $toolbox->getDatabaseAdapter();
		$db = $adapter->getCell("select database()");
		$fkCode = "fk".md5($table.$property1.$property2);
		$fks =  $adapter->getCell("
			SELECT count(*)
			FROM information_schema.KEY_COLUMN_USAGE
			WHERE TABLE_SCHEMA ='$db' AND TABLE_NAME ='$table' AND
			CONSTRAINT_NAME <>'PRIMARY' AND REFERENCED_TABLE_NAME is not null
		");
		
		if ($fks>0) return false;
		
		if (!$dontCache) self::$fkcache[ $fkCode ] = true;
		$columns = $writer->getColumns($table);
		if ($writer->code($columns[$property1])!==RedBean_QueryWriter_MySQL::C_DATATYPE_UINT32) {
			$writer->widenColumn($table, $property1, RedBean_QueryWriter_MySQL::C_DATATYPE_UINT32);
		}
		if ($writer->code($columns[$property2])!==RedBean_QueryWriter_MySQL::C_DATATYPE_UINT32) {
			$writer->widenColumn($table, $property2, RedBean_QueryWriter_MySQL::C_DATATYPE_UINT32);
		}
		$sql = "
			ALTER TABLE ".$writer->noKW($table)."
			ADD FOREIGN KEY($property1) references $table1(id) ON DELETE CASCADE;
		";
		$adapter->exec( $sql );
		$sql ="
			ALTER TABLE ".$writer->noKW($table)."
			ADD FOREIGN KEY($property2) references $table2(id) ON DELETE CASCADE
		";
		$adapter->exec( $sql );
		return true;
	}
	
	private static function constraintSQLite($toolbox, $table, $table1, $table2, $property1, $property2, $dontCache) {
		$writer = $toolbox->getWriter();
		$oodb = $toolbox->getRedBean();
		$adapter = $toolbox->getDatabaseAdapter();
		$fkCode = "fk".md5($table.$property1.$property2);
		$sql1 = "
			CREATE TRIGGER IF NOT EXISTS {$fkCode}a
				BEFORE DELETE ON $table1
				FOR EACH ROW BEGIN
					DELETE FROM $table WHERE  $table.$property1 = OLD.id;
				END;
		";
		$sql2 = "
			CREATE TRIGGER IF NOT EXISTS {$fkCode}b
				BEFORE DELETE ON $table2
				FOR EACH ROW BEGIN
					DELETE FROM $table WHERE $table.$property2 = OLD.id;
				END;
		";
		$adapter->exec($sql1);
		$adapter->exec($sql2);
		return true;
	}
}


abstract class RedBean_DomainObject {
	
	protected $tools;
	
	protected $redbean;
	
	protected $bean;
	
	protected $associationManager;
	
	protected $treeManager;
	
	public function __construct( $typeName = false ) {
	
		if (!$typeName) {
			
			$beanTypeName = get_class( $this );
			
			$a = explode( "\\" , $beanTypeName );
			$lastInNameSpace = array_pop( $a );
			
			$a = explode( "_" , $lastInNameSpace );
			$lastInPoormanNameSpace = array_pop( $a );
			$beanTypeName = $lastInPoormanNameSpace;
		}
		else {
			$beanTypeName = $typeName;
		}
		
		if ($beanTypeName && strlen($beanTypeName)>0) {
			
			$this->tools = RedBean_Setup::getToolBox();
			$this->redbean = $this->tools->getRedBean();
			
			$this->bean = $this->redbean->dispense( strtolower( $beanTypeName ) );
			
			$this->associationManager = new RedBean_AssociationManager($this->tools);
			$this->treeManager = new RedBean_TreeManager($this->tools);
		}
		else {
			throw new Exception("Invalid Domain Object TypeName");
		}
	}
	
	protected function associate(RedBean_DomainObject $other) {
		$this->associationManager->associate($this->bean, $other->bean);
	}
	
	protected function unassociate(RedBean_DomainObject $other) {
		$this->associationManager->unassociate($this->bean, $other->bean);
	}
	
	protected function related( $className, $constructorArg = null ) {
		$models = array();
		$model = new $className;
		$keys = $this->associationManager->related($this->bean, $model->getBeanType());
		foreach($keys as $key) {
			$modelItem = new $className($constructorArg);
			$modelItem->find( (int) $key );
			$models[$key] = $modelItem;
		}
		return $models;
	}
	
	protected function getBeanType() {
		return $this->bean->getMeta("type");
	}
	
	protected function clearRelations( $type ) {
		$this->associationManager->clearRelations($this->bean, $type);
	}
	
	protected function attach(RedBean_DomainObject $other) {
		$this->treeManager->attach($this->bean, $other->bean);
	}
	
	public function find( $id ) {
		$this->bean = $this->redbean->load( $this->bean->getMeta("type"), (int) $id );
	}
	
	
	public function save() {
		$this->redbean->store( $this->bean );
	}
	
	public function delete() {
		$this->redbean->trash( $this->bean );
	}
	
	public function getID() {
		$idField = $this->tools->getWriter()->getIDField( $this->bean->getMeta("type") );
		return $this->bean->$idField;
	}
	public function export() {
		return $this->bean;
	}
	public static function exportAll( $objects ) {
		$beans = array();
		foreach($objects as $object) {
			$beans[] = $object->export();
		}
		return $beans;
	}
	public function loadBean( RedBean_OODBBean $bean ) {
		$this->bean = $bean;
	}
}


class RedBean_Plugin_Optimizer extends RedBean_CompatManager implements RedBean_Plugin,RedBean_Observer {
	
	protected $supportedSystems = array(
		RedBean_CompatManager::C_SYSTEM_MYSQL => "5"
	);
	
	private $adapter;
	
	private $oodb;
	
	private $writer;
	
	public function __construct( RedBean_ToolBox $toolbox ) {
		$this->scanToolBox( $toolbox );
		$this->oodb = $toolbox->getRedBean();
		$this->adapter = $toolbox->getDatabaseAdapter();
		$this->writer = $toolbox->getWriter();
		
	}
	
	public function onEvent( $event , $bean ) {
		try{
		if ($event=="update") {
			$arr = $bean->export();
			unset($arr["id"]);
			if (count($arr)==0) return;
			$table = $this->adapter->escape($bean->getMeta("type"));
			$columns = array_keys($arr);
			
			$column = $this->adapter->escape($columns[ array_rand($columns) ]);
			$value = $arr[$column];
			$type = $this->writer->scanType($value);
			$fields = $this->writer->getColumns($table);
			if (!in_array($column,array_keys($fields))) return;
			$typeInField = $this->writer->code($fields[$column]);
			
			if ($type < $typeInField) {
				try{@$this->adapter->exec("alter table ".$this->writer->noKW($table)." drop __test");}catch(Exception $e){}
				
				$type = $this->writer->typeno_sqltype[$type];
				
				@$this->adapter->exec("alter table ".$this->writer->noKW($table)." add __test ".$type);
				
				@$this->adapter->exec("update ".$this->writer->noKW($table)." set __test=".$this->writer->noKW($column)."");
				$rows = $this->adapter->get("select ".$this->writer->noKW($column)." as a, __test as b from ".$this->writer->noKW($table));
				$diff = 0;
				foreach($rows as $row){ $diff += ($row["a"]!=$row["b"]); }
				if (!$diff) {
					
					@$this->adapter->exec("alter table ".$this->writer->noKW($table)." change ".$this->writer->noKW($column)." ".$this->writer->noKW($column)." ".$type);
				}
				
				@$this->adapter->exec("alter table ".$this->writer->noKW($table)." drop __test");
			}
			else {
				$this->MySQLSpecificColumns($table, $column, $fields[$column], $value);
			}
		}
		}catch(RedBean_Exception_SQL $e){
			
			
		}
	}
	
	public function MySQLSpecificColumns( $table, $column, $columnType, $value ) {
		
		$table = $this->adapter->escape($table);
		$column = $this->adapter->escape($column);
		
		if ($columnType!="datetime") {
			$pattern = "/^([0-9]{2,4})-([0-1][0-9])-([0-3][0-9]) (?:([0-2][0-9]):([0-5][0-9]):([0-5][0-9]))?$/";
			if (preg_match($pattern, $value)){
				
				$cnt = (int) $this->adapter->getCell("select count(*) as n from $table where
					$column regexp '[0-9]{4}-[0-1][0-9]-[0-3][0-9] [0-2][0-9]:[0-5][0-9]:[0-5][0-9]'
				");
				$total = (int) $this->adapter->getCell("SELECT count(*) FROM ".$this->writer->noKW($table));
				
				if ($total===$cnt) { 
					$this->adapter->exec("ALTER TABLE ".$this->writer->noKW($table)." change ".$this->writer->noKW($column)." ".$this->writer->noKW($column)." datetime ");
				}
			}
		}
	}
}

class RedBean_QueryWriter_NullWriter implements RedBean_QueryWriter {
	
	const C_DATATYPE_BOOL = 0;
	
	const C_DATATYPE_UINT8 = 1;
	
	const C_DATATYPE_UINT32 = 2;
	
	const C_DATATYPE_DOUBLE = 3;
	
	const C_DATATYPE_TEXT8 = 4;
	
	const C_DATATYPE_TEXT16 = 5;
	
	const C_DATATYPE_TEXT32 = 6;
	
	const C_DATATYPE_SPECIFIED = 99;
	
	public $createTableArgument = NULL;
	
	public $getColumnsArgument = NULL;
	
	public $scanTypeArgument = NULL;
	
	public $addColumnArguments = array();
	
	public $codeArgument = NULL;
	
	public $widenColumnArguments = array();
	
	public $updateRecordArguments = array();
	
	public $insertRecordArguments = array();
	
	public $selectRecordArguments = array();
	
	public $deleteRecordArguments = array();
	
	public $checkChangesArguments = array();
	
	public $addUniqueIndexArguments = array();
	
	public $selectByCritArguments = array();
	
	public $deleteByCrit = array();
	
	public $returnTables = array();
	
	public $returnGetColumns = array();
	
	public $returnScanType = 1;
	
	public $returnAddColumn = NULL;
	
	public $returnCode = NULL;
	
	public $returnWidenColumn = NULL;
	
	public $returnUpdateRecord = NULL;
	
	public $returnInsertRecord = NULL;
	
	public $returnSelectRecord = NULL;
	
	public $returnDeleteRecord = NULL;
	
	public $returnCheckChanges = NULL;
	
	public $returnDeleteByCrit = NULL;
	
	public$returnSelectByCrit = NULL;
	
	public $returnAddUniqueIndex = NULL;
	public function getTables(){ return $this->returnTables; }
	public function createTable( $table ){ $this->createTableArgument = $table; }
    public function getColumns( $table ){ $this->getColumnsArgument = $table; return $this->returnGetColumns; }
	public function scanType( $value ){ $this->scanTypeArgument = $value; return $this->returnScanType; }
    public function addColumn( $table, $column, $type ){
		$this->addColumnArguments = array( $table, $column, $type );
		return $this->returnAddColumn;
	}
	
    public function code( $typedescription ){ $this->codeArgument = $typedescription;
		return $this->returnCode;
	}
    public function widenColumn( $table, $column, $type ){
		$this->widenColumnArguments = array($table, $column, $type);
		return $this->returnWidenColumn;
	}
    public function updateRecord( $table, $updatevalues, $id){
		$this->updateRecordArguments = array($table, $updatevalues, $id);
		return $this->returnUpdateRecord;
	}
    public function insertRecord( $table, $insertcolumns, $insertvalues ){
		$this->insertRecordArguments = array( $table, $insertcolumns, $insertvalues );
		return $this->returnInsertRecord;
	}
    public function selectRecord($type, $ids){
		$this->selectRecordArguments = array($type, $ids);
		return $this->returnSelectRecord;
	}
	public function deleteRecord( $table, $id){
		$this->deleteRecordArguments = array($table, "id", $id);
		return $this->returnDeleteRecord;
	}
    public function checkChanges($type, $id, $logid){
		$this->checkChangesArguments = array($type, $id, $logid);
		return $this->returnCheckChanges;
	}
	public function addUniqueIndex( $table,$columns ){
		$this->addUniqueIndexArguments=array($table,$columns);
		return $this->returnAddUniqueIndex;
	}
	public function selectByCrit( $select, $table, $column, $value, $withUnion=false ){
		$this->selectByCritArguments=array($select, $table, $column, $value, $withUnion);
		return $this->returnSelectByCrit;
	}
	public function deleteByCrit( $table, $crits ){
		$this->deleteByCrit=array($table, $crits );
		return $this->returnDeleteByCrit;
	}
	public function getIDField( $type ) { return "id"; }
	public function noKW($str) { return $str; }
	public function sqlStateIn($state,$list) { return true; }
	
	public function reset() {
		$this->createTableArgument = NULL;
		$this->getColumnsArgument = NULL;
		$this->scanTypeArgument = NULL;
		$this->addColumnArguments = array();
		$this->codeArgument = NULL;
		$this->widenColumnArguments = array();
		$this->updateRecordArguments = array();
		$this->insertRecordArguments = array();
		$this->selectRecordArguments = array();
		$this->deleteRecordArguments = array();
		$this->checkChangesArguments = array();
		$this->addUniqueIndexArguments = array();
		$this->returnTables = array();
		$this->returnGetColumns = array();
		$this->returnScanType = 1;
		$this->returnAddColumn = NULL;
		$this->returnCode = NULL;
		$this->returnWidenColumn = NULL;
		$this->returnUpdateRecord = NULL;
		$this->returnInsertRecord = NULL;
		$this->returnSelectRecord = NULL;
		$this->returnDeleteRecord = NULL;
		$this->returnCheckChanges = NULL;
		$this->returnAddUniqueIndex = NULL;
	}
}

class RedBean_UnitOfWork {
	
	private $todoList = array();
	
	public function addWork( $tagName, $closure ) {
		if (strlen($tagName)>0) {
			if (!isset($this->todoList[$tagName])){
				
				$this->todoList[$tagName]=array();
			}
			$this->todoList[$tagName][] = $closure;
		}
	}
	
	public function doWork( $tagName ) {
		if (isset($this->todoList[$tagName])){
			foreach($this->todoList[$tagName] as $job) {
				$job();
			}
		}
	}
}


class RedBean_SimpleModel {
    
    protected $bean;
    
    public function loadBean( RedBean_OODBBean $bean ) {
        $this->bean = $bean;
    }
    
    public function __get( $prop ) {
        return $this->bean->$prop;
    }
    
    public function __set( $prop, $value ) {
        $this->bean->$prop = $value;
    }
}

class RedBean_ModelHelper implements RedBean_Observer {
    
    public function onEvent( $eventName, $bean ) {
       $className = "Model_".ucfirst( $bean->getMeta("type") );
       if (class_exists($className)) {
        $model = new $className;
        if ($model instanceof RedBean_SimpleModel) {
            $model->loadBean( $bean );
            if (method_exists($model, $eventName)){
                $model->$eventName();
            }
           }
       }
    }
}

class R {
	
	
	public static $toolbox;
	
	public static $redbean;
	
	public static $writer;
	
	public static $adapter;
	
	public static $treeManager;
	
	public static $associationManager;
        
	public static $linkManager;
	
	public static function setup( $dsn="sqlite:/tmp/red.db", $username=NULL, $password=NULL ) {
		RedBean_Setup::kickstart( $dsn, $username, $password );
		self::$toolbox = RedBean_Setup::getToolBox();
		self::$writer = self::$toolbox->getWriter();
		self::$adapter = self::$toolbox->getDatabaseAdapter();
		self::$redbean = self::$toolbox->getRedBean();
		self::$associationManager = new RedBean_AssociationManager( self::$toolbox );
		self::$treeManager = new RedBean_TreeManager( self::$toolbox );
                self::$linkManager = new RedBean_LinkManager( self::$toolbox );
                $helper = new RedBean_ModelHelper();
                self::$redbean->addEventListener("update", $helper );
                self::$redbean->addEventListener("open", $helper );
                self::$redbean->addEventListener("delete", $helper );
	}
	
	public static function store( RedBean_OODBBean $bean ) {
		return self::$redbean->store( $bean );
	}
        
        public static function freeze( $tf = true ) {
            self::$redbean->freeze( $tf );
        }
	
	public static function load( $type, $id ) {
		return self::$redbean->load( $type, $id );
	}
	
	public static function trash( RedBean_OODBBean $bean ) {
		return self::$redbean->trash( $bean );
	}
	
	public static function dispense( $type ) {
		return self::$redbean->dispense( $type );
	}
        
        public static function loadOrDispense( $type, $id = 0 ) {
            return ($id ? R::load($type,(int)$id) : R::dispense($type));
        }
        
	
	public static function associate( RedBean_OODBBean $bean1, RedBean_OODBBean $bean2 ) {
		return self::$associationManager->associate( $bean1, $bean2 );
	}
	
	public static function unassociate( RedBean_OODBBean $bean1, RedBean_OODBBean $bean2 ) {
		return self::$associationManager->unassociate( $bean1, $bean2 );
	}
	
	public static function related( RedBean_OODBBean $bean, $type ) {
		return self::$redbean->batch( $type, self::$associationManager->related( $bean, $type ));
	}
	
	public static function clearRelations( RedBean_OODBBean $bean, $type ) {
		return self::$associationManager->clearRelations( $bean, $type );
	}
	
	public static function attach( RedBean_OODBBean $parent, RedBean_OODBBean $child ) {
		return self::$treeManager->attach( $parent, $child );
	}
        
        public static function link( RedBean_OODBBean $bean1, RedBean_OODBBean $bean2 ) {
                return self::$linkManager->link( $bean1, $bean2 );
        }
        
        public static function getBean( RedBean_OODBBean $bean, $typeName ) {
                return self::$linkManager->getBean($bean, $typeName );
        }
        
        public static function getKey( RedBean_OODBBean $bean, $typeName ) {
                return self::$linkManager->getKey($bean, $typeName );
        }
        
        public static function breakLink( RedBean_OODBBean $bean, $typeName ) {
            return self::$linkManager->breakLink( $bean, $typeName );
        }
	
	public static function children( RedBean_OODBBean $parent ) {
		return self::$treeManager->children( $parent );
	}
        public static function getParent( RedBean_OODBBean $bean ) {
                return self::$treeManager->getParent( $bean );
        }
	
	
	public static function find( $type, $where="1", $values=array() ) {
		return Finder::where( $type, $where, $values );
	}
        
        public static function batch( $type, $ids ) {
            return self::$redbean->batch($type, $ids);
        }
        
        public static function lst( $type,$prop,$where=" 1 " ) {
            $list = self::find($type,$where);
            $listItems = array();
            foreach($list as $id=>$item) {
                $listItems[] = $item->$prop;
            }
            return $listItems;
        }
        
        public static function exec( $sql, $values=array() ) {
            return self::secureExec(function($sql, $values){return R::$adapter->exec( $sql, $values );}, NULL,$sql, $values );
        }
        
        public static function getAll( $sql, $values=array() ) {
            return self::secureExec(function($sql, $values){return R::$adapter->get( $sql, $values );}, array(), $sql, $values);
        }
        
        public static function getCell( $sql, $values=array() ) {
            return self::secureExec(function($sql, $values){return R::$adapter->getCell( $sql, $values );}, NULL, $sql, $values);
        }
        
        public static function getRow( $sql, $values=array() ) {
            return self::secureExec(function($sql, $values){return R::$adapter->getRow( $sql, $values );}, array(),$sql, $values);
        }
        
        public static function getCol( $sql, $values=array() ) {
            return self::secureExec(function($sql, $values){return R::$adapter->getCol( $sql, $values );}, array(),$sql, $values);
        }
        
        private static function secureExec( $func, $default=NULL, $sql, $values ) {
            if (!self::$redbean->isFrozen()) { 
                try{ $rs = $func($sql,$values);  }catch(RedBean_Exception_SQL $e) { 
                    if(self::$writer->sqlStateIn($e->getSQLState(),
                    array(
			RedBean_QueryWriter::C_SQLSTATE_NO_SUCH_COLUMN,
			RedBean_QueryWriter::C_SQLSTATE_NO_SUCH_TABLE)
                    )){
                       return $default;
                    }
                    else {
                        throw $e;
                    }
                    
                }
                return $rs;
            }
        }
}
